@@ -9,9 +9,9 @@ from typing import Any
 import click
 
 from pglift import patroni
-from pglift.models import system
+from pglift.models import Instance
 from pglift.patroni import impl
-from pglift.patroni import register_if as register_if  # noqa: F401
+from pglift.patroni import register_if as register_if
 
 from . import _site, hookimpl
 from .util import Group, instance_identifier_option, pass_instance
@@ -25,7 +25,7 @@ def cli(**kwargs: Any) -> None:
 
 @cli.command("logs")
 @pass_instance
-def logs(instance: system.Instance) -> None:
+def logs(instance: Instance) -> None:
     """Output Patroni logs."""
     settings = patroni.get_settings(_site.SETTINGS)
     for line in impl.logs(instance.qualname, settings):
