@@ -7,12 +7,12 @@ from __future__ import annotations
 import click
 
 from pglift import pgbackrest, types
-from pglift.models import system
+from pglift.models import Instance
 from pglift.pgbackrest import repo_path
-from pglift.pgbackrest.repo_path import register_if as register_if  # noqa: F401
+from pglift.pgbackrest.repo_path import register_if as register_if
 
 from .. import _site, hookimpl
-from ..util import Command, Obj, async_command, instance_identifier
+from ..util import Command, Obj, async_command, audit, instance_identifier
 
 
 @click.command("backup", cls=Command)
@@ -27,11 +27,11 @@ from ..util import Command, Obj, async_command, instance_identifier
 @click.pass_obj
 @async_command
 async def instance_backup(
-    obj: Obj, instance: system.Instance, backup_type: types.BackupType
+    obj: Obj, instance: Instance, backup_type: types.BackupType
 ) -> None:
     """Back up PostgreSQL INSTANCE"""
     settings = pgbackrest.get_settings(_site.SETTINGS)
-    with obj.lock:
+    with obj.lock, audit():
         await repo_path.backup(instance, settings, type=backup_type)
 
 
