@@ -6,25 +6,26 @@ from __future__ import annotations
 
 import click
 
-from pglift.cmd import execute_program
+from pglift import postgresql
 from pglift.exceptions import InstanceNotFound
-from pglift.models import system
+from pglift.models import PostgreSQLInstance
+from pglift.system import cmd
 
 from . import _site
 
 
 def instance_from_qualname(
-    context: click.Context, param: click.Parameter, value: str
-) -> system.PostgreSQLInstance:
+    context: click.Context, _param: click.Parameter, value: str
+) -> PostgreSQLInstance:
     try:
-        return system.PostgreSQLInstance.from_qualname(value, _site.SETTINGS)
+        return PostgreSQLInstance.from_qualname(value, _site.SETTINGS)
     except (ValueError, InstanceNotFound) as e:
         raise click.BadParameter(str(e), context) from None
 
 
 @click.command("postgres", hidden=True)
 @click.argument("instance", callback=instance_from_qualname)
-def cli(instance: system.Instance) -> None:
+def cli(instance: PostgreSQLInstance) -> None:
     """Start postgres for specified INSTANCE, identified as <version>-<name>."""
-    cmd = [str(instance.bindir / "postgres"), "-D", str(instance.datadir)]
-    execute_program(cmd)
+    args = [str(postgresql.bindir(instance) / "postgres"), "-D", str(instance.datadir)]
+    cmd.execute_program(args)
