@@ -6,43 +6,47 @@ from __future__ import annotations
 
 import abc
 import asyncio
-import enum
+import inspect
 import json
 import logging
+import logging.handlers
 import os
-import pathlib
+import shlex
+import sys
 import tempfile
 import time
 import typing
-from collections.abc import Coroutine, Iterable, Iterator, Sequence
-from contextlib import contextmanager
-from functools import cache, cached_property, singledispatch, wraps
-from typing import Any, Callable, TypeVar
+from collections.abc import Callable, Coroutine, Iterable, Iterator, Sequence
+from contextlib import contextmanager, nullcontext
+from datetime import timedelta
+from functools import cache, cached_property, partial, singledispatch, wraps
+from pathlib import Path
+from typing import IO, Any, Literal, ParamSpec, TypedDict, TypeVar
 
 import click
 import filelock
+import humanize
 import psycopg
 import pydantic
 import pydantic_core
 import rich
 import rich.prompt
+import yaml
 from click.shell_completion import CompletionItem
 from rich.console import Console
 from rich.table import Table
 
-from pglift import exceptions, install, instances, task
-from pglift._compat import ParamSpec
-from pglift.models import helpers, system
-from pglift.settings import Settings
-from pglift.settings._postgresql import PostgreSQLVersion
-from pglift.task import Displayer
-from pglift.types import AutoStrEnum, ByteSizeType
+from pglift import exceptions, instances, system
+from pglift.models import Instance, PostgreSQLInstance, helpers, interface
+from pglift.settings import PostgreSQLVersion, Settings
+from pglift.system import install
+from pglift.types import ByteSizeType
 
-from . import _site, model
+from . import __name__ as pkgname
+from . import _site, logger, loggers
+from ._settings import CLISettings
 from .console import console
 
-logger = logging.getLogger("pglift")
-
 
 def model_dump(
     m: pydantic.BaseModel, by_alias: bool = True, **kwargs: Any
@@ -51,7 +55,7 @@ def model_dump(
 
 
 @singledispatch
-def prettify(value: Any, annotations: Sequence[Any] = ()) -> str:
+def prettify(value: Any, *args: Any) -> str:
     """Prettify a value."""
     return str(value)
 
@@ -66,7 +70,7 @@ def _(value: int, annotations: Sequence[Any] = ()) -> str:
 
 
 @prettify.register(list)
-def _(value: list[Any], annotations: Sequence[Any] = ()) -> str:
+def _(value: list[Any], *args: Any) -> str:
     """Prettify a List value"""
     return ", ".join(str(x) for x in value)
 
@@ -78,13 +82,13 @@ def _(value: set[Any], annotations: Sequence[Any] = ()) -> str:
 
 
 @prettify.register(type(None))
-def _(value: None, annotations: Sequence[Any] = ()) -> str:
+def _(*args: Any) -> str:
     """Prettify a None value"""
     return ""
 
 
 @prettify.register(dict)
-def _(value: dict[str, Any], annotations: Sequence[Any] = ()) -> str:
+def _(value: dict[str, Any], *args: Any) -> str:
     """Prettify a Dict value"""
 
     def prettify_dict(
@@ -113,7 +117,6 @@ def print_table_for(
     **kwargs: Any,
 ) -> None:
     """Render a list of items as a table."""
-    table = None
     headers: list[str] = []
     rows = []
     for item in items:
@@ -136,7 +139,13 @@ def print_table_for(
         rows.append(row)
     if not rows:
         return
-    table = Table(*headers, title=title, **kwargs)
+    table = Table(title=title, **kwargs)
+    # https://github.com/Textualize/rich/issues/3761
+    overflow: Literal["ellipsis", "fold"] = (
+        "fold" if console.options.ascii_only else "ellipsis"
+    )
+    for header in headers:
+        table.add_column(header, overflow=overflow)
     for row in rows:
         table.add_row(*row)
     console.print(table)
@@ -144,7 +153,15 @@ def print_table_for(
 
 def print_json_for(data: Any, *, console: Console = console) -> None:
     """Render `data` as JSON."""
-    console.print_json(data=pydantic_core.to_jsonable_python(data))
+    console.print_json(data=pydantic_core.to_jsonable_python(data, by_alias=True))
+
+
+def print_result_diff(
+    r: interface.ApplyResult, /, *, console: Console = console
+) -> None:
+    if r.diff is not None:
+        for diffitem in r.diff:
+            console.print(diffitem)
 
 
 P = ParamSpec("P")
@@ -152,7 +169,7 @@ P = ParamSpec("P")
 
 def print_schema(
     context: click.Context,
-    param: click.Parameter,
+    _param: click.Parameter,
     value: bool,
     *,
     model: type[pydantic.BaseModel],
@@ -166,7 +183,7 @@ def print_schema(
 
 def print_argspec(
     context: click.Context,
-    param: click.Parameter,
+    _param: click.Parameter,
     value: bool,
     *,
     model: type[pydantic.BaseModel],
@@ -186,19 +203,37 @@ def pass_instance(f: Callable[P, None]) -> Callable[P, None]:
     def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
         context = click.get_current_context()
         instance = context.obj.instance
-        assert isinstance(instance, system.Instance), instance
+        assert isinstance(instance, Instance), instance
         context.invoke(f, instance, *args, **kwargs)
 
     return wrapper
 
 
-def get_instance(name: str, version: str | None, settings: Settings) -> system.Instance:
-    """Return an Instance from name/version, possibly guessing version if unspecified."""
+def pass_postgresql_instance(f: Callable[P, None]) -> Callable[P, None]:
+    """Command decorator passing PostgreSQL 'instance' bound to
+    click.Context's object.
+    """
+
+    @wraps(f)
+    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
+        context = click.get_current_context()
+        instance = context.obj.postgresql_instance
+        assert isinstance(instance, PostgreSQLInstance), instance
+        context.invoke(f, instance, *args, **kwargs)
+
+    return wrapper
+
+
+def get_postgresql_instance(
+    name: str, version: PostgreSQLVersion | None, settings: Settings
+) -> PostgreSQLInstance:
+    """Return a PostgreSQLInstance from name/version, possibly guessing version if unspecified."""
     if version is None:
         found = None
-        for version in PostgreSQLVersion:
+        for v in settings.postgresql.versions:
+            version = v.version
             try:
-                instance = system.Instance.system_lookup((name, version, settings))
+                instance = PostgreSQLInstance.system_lookup(name, version, settings)
             except exceptions.InstanceNotFound:
                 logger.debug("instance '%s' not found in version %s", name, version)
             else:
@@ -215,31 +250,39 @@ def get_instance(name: str, version: str | None, settings: Settings) -> system.I
         raise click.BadParameter(f"instance {name!r} not found")
 
     try:
-        return system.Instance.system_lookup((name, version, settings))
+        return PostgreSQLInstance.system_lookup(name, version, settings)
     except Exception as e:
         raise click.BadParameter(str(e)) from None
 
 
-def nameversion_from_id(instance_id: str) -> tuple[str, str | None]:
+def get_instance(
+    name: str, version: PostgreSQLVersion | None, settings: Settings
+) -> Instance:
+    """Return an Instance from name/version, possibly guessing version if unspecified."""
+    pg_instance = get_postgresql_instance(name, version, settings)
+    return Instance.from_postgresql(pg_instance)
+
+
+def nameversion_from_id(instance_id: str) -> tuple[str, PostgreSQLVersion | None]:
     version = None
     try:
         version, name = instance_id.split("/", 1)
     except ValueError:
         name = instance_id
-    return name, version
+    return name, typing.cast(PostgreSQLVersion, version)
 
 
-def instance_lookup(
+def postgresql_instance_lookup(
     context: click.Context, param: click.Parameter, value: None | str | tuple[str]
-) -> system.Instance | tuple[system.Instance, ...]:
-    """Return one or more system.Instance, possibly guessed if there is only
-    one on system, depending on 'param' variadic flag (nargs).
+) -> PostgreSQLInstance | tuple[PostgreSQLInstance, ...]:
+    """Return one or more PostgreSQLInstance, possibly guessed if there
+    is only one on system, depending on 'param' variadic flag (nargs).
     """
 
     settings = _site.SETTINGS
 
-    def guess() -> tuple[str, str | None]:
-        """Return (name, version) of the instance found on system, if there's
+    def guess() -> PostgreSQLInstance:
+        """Return the PostgreSQLInstance found on system, if there's
         only one, or fail.
         """
         try:
@@ -248,44 +291,52 @@ def instance_lookup(
             raise click.UsageError(
                 f"argument {param.get_error_hint(context)} is required."
             ) from None
-        return i.name, i.version
+        return i
 
-    if context.params.get("all_instances"):
-        return tuple(
-            get_instance(i.name, i.version, i._settings)
-            for i in instances.system_list(settings)
-        )
+    if context.params.get("_all_instances"):
+        return tuple(instances.system_list(settings))
 
     if param.nargs == 1:
         if value is None:
-            name, version = guess()
+            return guess()
         else:
             assert isinstance(value, str)
             name, version = nameversion_from_id(value)
-        return get_instance(name, version, settings)
+            return get_postgresql_instance(name, version, settings)
 
     elif param.nargs == -1:
         assert isinstance(value, tuple)
         if value:
             return tuple(
-                get_instance(*nameversion_from_id(item), settings) for item in value
+                get_postgresql_instance(*nameversion_from_id(item), settings)
+                for item in value
             )
         else:
-            name, version = guess()
-            return (get_instance(name, version, settings),)
+            return (guess(),)
 
     else:
         raise AssertionError(f"unexpected nargs={param.nargs}")
 
 
+def instance_lookup(
+    context: click.Context, param: click.Parameter, value: None | str | tuple[str]
+) -> Instance | tuple[Instance, ...]:
+    """Return one or more Instance, possibly guessed if there is only
+    one on system, depending on 'param' variadic flag (nargs).
+    """
+    pg_value = postgresql_instance_lookup(context, param, value)
+    if isinstance(pg_value, tuple):
+        return tuple(Instance.from_postgresql(i) for i in pg_value)
+    return Instance.from_postgresql(pg_value)
+
+
 def instance_bind_context(
     context: click.Context, param: click.Parameter, value: str | None
 ) -> None:
-    """Bind instance specified as -i/--instance to context's object, possibly
-    guessing from available instance if there is only one.
+    """Bind PostgreSQL instance specified as -i/--instance to context's
+    object, possibly guessing from available instance if there is only one.
     """
     obj: Obj = context.obj
-    version: str | None
     if value is None:
         values = list(instances.system_list(_site.SETTINGS))
         if not values:
@@ -295,16 +346,16 @@ def instance_bind_context(
             option = param.get_error_hint(context)
             obj._instance = f"several instances found; option {option} is required."
             return
-        (i,) = values
-        name, version = i.name, i.version
+        (pg_instance,) = values
     else:
-        name, version = nameversion_from_id(value)
-    instance = get_instance(name, version, _site.SETTINGS)
-    obj._instance = instance
+        pg_instance = get_postgresql_instance(
+            *nameversion_from_id(value), _site.SETTINGS
+        )
+    obj._instance = pg_instance
 
 
 def _list_instances(
-    context: click.Context, param: click.Parameter, incomplete: str
+    _context: click.Context, _param: click.Parameter, incomplete: str
 ) -> list[CompletionItem]:
     """Shell completion function for instance identifier <name> or <version>/<name>."""
     out = []
@@ -327,21 +378,27 @@ def _list_instances(
 F = TypeVar("F", bound=Callable[..., Any])
 
 
-def instance_identifier(nargs: int = 1, required: bool = False) -> Callable[[F], F]:
+def _instance_identifier(
+    callback: Callable[[click.Context, click.Parameter, Any], Any],
+    nargs: int = 1,
+    required: bool = False,
+) -> Callable[[F], F]:
     def decorator(fn: F) -> F:
         command = click.argument(
             "instance",
             nargs=nargs,
             required=required,
-            callback=instance_lookup,
+            callback=callback,
             shell_complete=_list_instances,
         )(fn)
-        assert command.__doc__
-        command.__doc__ += (
-            "\n\n    INSTANCE identifies target instance as <version>/<name> where the "
+        doc = inspect.getdoc(command)
+        assert doc
+        doc += (
+            "\n\nINSTANCE identifies target instance as <version>/<name> where the "
             "<version>/ prefix may be omitted if there is only one instance "
             "matching <name>."
         )
+        command.__doc__ = doc
         if not required:
             command.__doc__ += " Required if there is more than one instance on system."
         return command
@@ -349,6 +406,11 @@ def instance_identifier(nargs: int = 1, required: bool = False) -> Callable[[F],
     return decorator
 
 
+postgresql_instance_identifier = partial(
+    _instance_identifier, postgresql_instance_lookup
+)
+instance_identifier = partial(_instance_identifier, instance_lookup)
+
 instance_identifier_option = click.option(
     "-i",
     "--instance",
@@ -364,26 +426,77 @@ instance_identifier_option = click.option(
 )
 
 
-class OutputFormat(AutoStrEnum):
-    """Output format"""
+def yaml_load(
+    _ctx: click.Context, _param: click.Parameter, value: IO[str]
+) -> dict[str, Any]:
+    try:
+        data = yaml.safe_load(value)
+    except yaml.YAMLError as e:
+        raise click.BadParameter(f"invalid YAML: {e}") from e
+    if not isinstance(data, dict):
+        raise click.BadParameter(f"invalid YAML: expecting an object, got {type(data)}")
+    if "name" not in data:
+        raise click.BadParameter("invalid YAML: missing required 'name' field")
+    return data
 
-    json = enum.auto()
+
+manifest_option = click.option(
+    "-f",
+    "--file",
+    "data",
+    type=click.File("r"),
+    metavar="MANIFEST",
+    required=True,
+    callback=yaml_load,
+)
+
+
+class ManifestData(TypedDict, total=False):
+    name: str
+
+
+OutputFormat = Literal["json"]
+
+
+def set_output_format(
+    ctx: click.Context, _param: click.Parameter, value: OutputFormat
+) -> OutputFormat:
+    assert ctx.obj.output_format is None
+    ctx.obj.output_format = value
+    return value
 
 
 output_format_option = click.option(
     "-o",
     "--output-format",
-    type=click.Choice(model.choices_from_enum(OutputFormat), case_sensitive=False),
+    type=click.Choice(typing.get_args(OutputFormat), case_sensitive=False),
+    callback=set_output_format,
     help="Specify the output format.",
 )
 
+diff_options = {
+    "unified": click.option(
+        "--diff",
+        "diff_format",
+        flag_value="unified",
+        help="Include differences resulting from applied changes in returned result.",
+    ),
+    "ansible": click.option(
+        "--ansible-diff",
+        "diff_format",
+        flag_value="ansible",
+        help="Include differences resulting from applied changes in returned result using Ansible's diff format.",
+        hidden=True,
+    ),
+}
+
 dry_run_option = click.option(
-    "--dry-run", is_flag=True, help="Only validate input data."
+    "--dry-run", is_flag=True, help="Simulate change operations."
 )
 
 
 def validate_foreground(
-    context: click.Context, param: click.Parameter, value: bool
+    _context: click.Context, _param: click.Parameter, value: bool
 ) -> bool:
     if _site.SETTINGS.service_manager == "systemd" and value:
         raise click.BadParameter("cannot be used with systemd")
@@ -399,29 +512,106 @@ foreground_option = click.option(
 
 
 @contextmanager
-def command_logging(logdir: pathlib.Path) -> Iterator[None]:
-    logdir_created = False
-    logfilename = f"{time.time()}.log"
-    logfile = logdir / logfilename
+def system_configure(*, dry_run: bool) -> Iterator[None]:
+    with system.configure(dry_run=dry_run):
+        yield
+    if dry_run:
+        click.echo("DRY RUN: no changes made", err=True)
+
+
+@contextmanager
+def audit(
+    command: Sequence[str] = sys.argv,
+    settings: CLISettings | None = None,
+    *,
+    dry_run: bool = False,
+) -> Iterator[None]:
+    """Context manager handling log messages to the audit file, if configured
+    in site settings.
+    """
+    if settings is None:
+        settings = _site.SETTINGS.cli
+    if (audit_settings := settings.audit) is None:
+        yield None
+        return
+
+    audit_file = audit_settings.path
+    if not audit_file.parent.exists():
+        logger.debug("creating audit file parent directory")
+        audit_file.parent.mkdir(parents=True, exist_ok=True)
+    logger.debug("using audit file %s", audit_file)
+
+    handler = logging.handlers.WatchedFileHandler(audit_file)
+    handler.setLevel(logging.DEBUG)
+    formatter = logging.Formatter(
+        fmt=audit_settings.log_format, datefmt=audit_settings.date_format
+    )
+    handler.setFormatter(formatter)
+
+    loggrs = [logging.getLogger(n) for n in loggers]
+    for loggr in loggrs:
+        loggr.addHandler(handler)
+
+    # The audit logger, as defined here, is only meant to emit start/end
+    # messages below; and we avoid them to get propagated to higher logger.
+    audit_logger = logging.getLogger(pkgname).getChild("audit")
+    audit_logger.propagate = False
+    audit_logger.addHandler(handler)
+    audit_logger.setLevel(logging.DEBUG)
+
+    msg = "command: %s"
+    if dry_run:
+        msg += " (DRY RUN)"
+    audit_logger.info(msg, shlex.join(command))
+    started_at = time.monotonic()
+
     try:
-        if not logdir.exists():
-            logdir.mkdir(parents=True)
-            logdir_created = True
-    except OSError:
-        # Might be, e.g. PermissionError, if log file path is not writable.
-        logfile = pathlib.Path(
-            tempfile.NamedTemporaryFile(prefix="pglift", suffix=logfilename).name
+        yield None
+    except Exception as exc:
+        if isinstance(exc, exceptions.Cancelled):
+            msg, level = "command cancelled (%s)", logging.WARNING
+        else:
+            msg, level = "command failed (%s)", logging.ERROR
+        raise
+    else:
+        msg, level = "command completed (%s)", logging.INFO
+    finally:
+        # Note: by removing the audit handler from loggers managed above, we
+        # avoid termination messages (typically those emitted in
+        # Command.invoke()) to be emitted in this handler.
+        # If this appears to be a bad idea, removeHandler() should be invoked
+        # through context.call_on_close().
+        for loggr in loggrs:
+            loggr.removeHandler(handler)
+        elapsed = humanize.precisedelta(
+            timedelta(seconds=time.monotonic() - started_at)
         )
+        audit_logger.log(level, msg, elapsed)
+        audit_logger.removeHandler(handler)
+        handler.close()
+
+
+@contextmanager
+def command_logging() -> Iterator[None]:
+    """Log any unexpected error, i.e. not one of our exceptions or one we
+    explicitly raised, to a debug file.
+
+    This only makes sense when not in --debug mode; otherwise, the bare
+    exception will propagate to stderr.
+    """
+    logfilename = f"{time.time()}.log"
+    logfile = Path(
+        tempfile.NamedTemporaryFile(prefix="pglift-", suffix="-" + logfilename).name
+    )
+    logger.debug("debug logging at %s", logfile)
     handler = logging.FileHandler(logfile)
     formatter = logging.Formatter(
         fmt="%(levelname)-8s - %(asctime)s - %(name)s:%(filename)s:%(lineno)d - %(message)s",
         datefmt="%Y-%m-%d %H:%M:%S",
     )
     handler.setFormatter(formatter)
-    logger.addHandler(handler)
-    if logdir_created:
-        logger.debug("created log directory %s", logdir)
-    logger.debug("logging command at %s", logfile)
+    for name in loggers:
+        logging.getLogger(name).addHandler(handler)
     keep_logfile = False
     try:
         yield None
@@ -429,7 +619,13 @@ def command_logging(logdir: pathlib.Path) -> Iterator[None]:
         raise
     except Exception:
         keep_logfile = True
-        logger.exception("an unexpected error occurred")
+        msg = "an unexpected error occurred"
+        logger.error(msg)
+        # Only log the traceback to the log file; for this, instantiate a
+        # specific logger (unrelated to application loggers hierarchy).
+        flogger = logging.getLogger(f"{pkgname}_debug")
+        flogger.addHandler(handler)
+        flogger.exception(msg)
         raise click.ClickException(
             "an unexpected error occurred, this is probably a bug; "
             f"details can be found at {logfile}"
@@ -437,14 +633,6 @@ def command_logging(logdir: pathlib.Path) -> Iterator[None]:
     finally:
         if not keep_logfile:
             os.unlink(logfile)
-            if logdir_created and next(logdir.iterdir(), None) is None:
-                logger.debug("removing log directory %s", logdir)
-                logdir.rmdir()
-
-
-class LogDisplayer:
-    def handle(self, msg: str) -> None:
-        logger.info(msg)
 
 
 class InteractiveUserInterface:
@@ -467,16 +655,13 @@ class Obj:
 
     # Set in commands taking a -i/--instance option through
     # instance_identifier_option decorator's callback.
-    _instance: str | system.Instance
+    _instance: str | PostgreSQLInstance
 
     def __init__(
-        self,
-        *,
-        displayer: Displayer | None = None,
-        debug: bool = False,
+        self, *, debug: bool = False, output_format: OutputFormat | None = None
     ) -> None:
-        self.displayer = displayer
         self.debug = debug
+        self.output_format = output_format
 
     @cached_property
     def lock(self) -> filelock.FileLock:
@@ -486,16 +671,20 @@ class Obj:
         return filelock.FileLock(lockfile, timeout=0)
 
     @cached_property
-    def instance(self) -> system.Instance:
-        if isinstance(self._instance, system.Instance):
+    def postgresql_instance(self) -> PostgreSQLInstance:
+        if isinstance(self._instance, PostgreSQLInstance):
             return self._instance
         else:
             assert isinstance(self._instance, str)
             raise click.UsageError(self._instance)
 
+    @cached_property
+    def instance(self) -> Instance:
+        return Instance.from_postgresql(self.postgresql_instance)
+
 
 def async_command(
-    callback: Callable[P, Coroutine[None, None, None]]
+    callback: Callable[P, Coroutine[None, None, None]],
 ) -> Callable[P, None]:
     @wraps(callback)
     def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
@@ -507,17 +696,18 @@ def async_command(
 class Command(click.Command):
     def invoke(self, context: click.Context) -> Any:
         obj: Obj = context.obj
-        displayer = obj.displayer
-        with command_logging(_site.SETTINGS.cli.logpath):
+        with command_logging() if not obj.debug else nullcontext(None):
             try:
-                with task.displayer_installed(displayer):
-                    return super().invoke(context)
+                return super().invoke(context)
             except filelock.Timeout:
                 raise click.ClickException("another operation is in progress") from None
             except exceptions.Cancelled as e:
                 logger.warning(str(e))
                 raise click.Abort from None
-            except (exceptions.ValidationError, pydantic.ValidationError) as e:
+            except pydantic.ValidationError as e:
+                logger.debug("a validation error occurred", exc_info=obj.debug)
+                if context.obj.output_format == "json":
+                    console.print_json(e.json(include_url=False, include_context=False))
                 raise click.ClickException(str(e)) from None
             except exceptions.Error as e:
                 logger.debug("an internal error occurred", exc_info=obj.debug)
@@ -529,10 +719,19 @@ class Command(click.Command):
                         msg += f"\n{e.stdout}"
                 raise click.ClickException(msg) from None
             except psycopg.DatabaseError as e:
-                logger.debug("a database error occurred", exc_info=True)
+                logger.debug(
+                    "a database error occurred: %s (SQLSTATE=%s)",
+                    e,
+                    e.sqlstate,
+                    exc_info=obj.debug,
+                )
                 raise click.ClickException(str(e).strip()) from None
 
 
+def is_root() -> bool:
+    return os.getuid() == 0
+
+
 class Group(click.Group):
     command_class = Command
     group_class = type
@@ -543,10 +742,13 @@ class Group(click.Group):
         super().add_command(command, name)
 
     def invoke(self, ctx: click.Context) -> Any:
-        if not install.check(_site.SETTINGS):
-            raise click.ClickException(
-                "broken installation; did you run 'site-configure install'?",
-            )
+        if set(ctx.help_option_names) - set(ctx.args):
+            if is_root():
+                raise click.ClickException("pglift cannot be used as root")
+            if not install.check(_site.SETTINGS):
+                raise click.ClickException(
+                    "broken installation; did you run 'site-configure install'?",
+                )
         return super().invoke(ctx)
 
 
